@@ -260,11 +260,12 @@ class PGPMIME(KryptoMIME):
 
     def _check_signatures(self,payload,signatures,rawmail,strict=True):
         "returns mail without signature, whether it was signed and with which keyids"
-        from .mail import protect_mail
+        from .mail import protect_mail, fix_lines
         results, key_ids = [], []
         signed, fmt = False, False
         for signature in signatures:
-            for ending in (None,'\n','\r\n'):
+            if signature: signature = self._fix_quoting(signature)
+            for ending in (None,'\r\n','\n'):
                 if ending:
                     tmp = fix_lines(payload,ending=ending)
                     if tmp==payload: continue
@@ -272,15 +273,14 @@ class PGPMIME(KryptoMIME):
                 if signature is None:
                     result = self.verify_str(tmp)
                 else:
-                    signature = self._fix_quoting(signature)
                     result = self.verify_str(tmp, signature)
                 if result.key_id: signed = True
                 if fmt or strict: break # already found format
-                if result: # found valid
+                if result.valid: # found valid
                     if not fmt and ending:
                         payload = tmp # correct format
                         rawmail = protect_mail(rawmail,ending=ending,sevenbit=False)
-                        fmt = True
+                    fmt = True
                     break
             if not result: continue # no valid signature
             key_ids.append(result.key_id)
@@ -406,7 +406,8 @@ class PGPMIME(KryptoMIME):
         if not 'default_key' in kwargs:
             import email.utils
             sender = mail.get('from', [])
-            if sender: kwargs['default_key'] = email.utils.parseaddr(sender)[1]
+            sender = email.utils.parseaddr(sender)[1]
+            kwargs['default_key'] = sender
         plaintext, submsg = self._plaintext(mail, inline)
         # Generate signature, report errors
         if not mail.is_multipart() and inline:
@@ -422,7 +423,7 @@ class PGPMIME(KryptoMIME):
             mail.set_param('x-action','pgp-signed')
             if verify:
                 vresult = self.verify_str(signature)
-                if not vresult: return None, result
+                if not vresult.valid: return None, result
         else:
             # workaround to preserve header order
             tmp = Message()
@@ -444,19 +445,21 @@ class PGPMIME(KryptoMIME):
             mail.attach(submsg)
             if verify:
                 vresult = self.verify_str(mail.get_payload(0).as_string(),signature)
-                if not vresult: return None, result
+                if not vresult.valid: return None, result
         return mail, result
 
-    def encrypt(self, mail, sign=True, inline = False, verify=False, **kwargs):
+    def encrypt(self, mail, sign=True, inline=False, toself=True, verify=False, **kwargs):
         """Encrypts an email for the recipients in To/CC.
         
         :param mail: A string or email
         :type mail: string or Message object
-        :param sign: Whether to sign the mail with the sender's (From) signature
+        :param sign: Whether to sign the mail with the sender's (From) signature.
         :type sign: bool
         :param inline: Whether to use the PGP inline format for messages with attachments, i.e. no multipart.
         :type inline: bool
-        :param verify: Whether to verify the encrypted mail immediately
+        :param toself: Whether to add sender to the recipients for self-decryption.
+        :type toself: bool
+        :param verify: Whether to verify the encrypted mail immediately (toself must be enabled or recipient key be known).
         :type verify: bool
         :returns: The encrypted email (None if it fails) and the encryption details.
         :rtype: (Message,dict)
@@ -467,12 +470,13 @@ class PGPMIME(KryptoMIME):
         from .mail import _mail_addreplace_header
         if type(mail)==str: mail = Parser().parsestr(mail)
         elif not isinstance(mail,Message): return None, None
-        if sign and not 'default_key' in kwargs:
-            sender = mail.get('from', [])
-            if sender: kwargs['default_key'] = email.utils.parseaddr(sender)[1]
+        sender = mail.get('from', [])
+        sender = email.utils.parseaddr(sender)[1]
         tos = mail.get_all('to', [])
         ccs = mail.get_all('cc', [])
         recipients = [email.utils.formataddr(to) for to in email.utils.getaddresses(tos + ccs)]
+        if toself and not sender in recipients: recipients.append(sender)
+        if sign and not 'default_key' in kwargs: kwargs['default_key'] = sender
         plaintext, submsg = self._plaintext(mail, inline, protect=False)
         # Do encryption, report errors
         kwargs['sign'] = sign
@@ -481,9 +485,11 @@ class PGPMIME(KryptoMIME):
         if not result: return None, result
         ciphertext = str(result)
         if verify:
+            del kwargs['sign']
+            del kwargs['armor']
+            if not 'default_key' in kwargs: kwargs['default_key'] = sender
             vresult = self.decrypt_str(ciphertext, **kwargs)
-            if not vresult: return None, result
-            if sign and not result.key_id: return None, result
+            if not vresult.ok or (sign and not vresult.valid): return None, result
         # Compile encrypted message
         encmail = HeaderParser().parsestr(mail.as_string())
         _mail_addreplace_header(encmail,'Content-Transfer-Encoding','7bit')
@@ -511,8 +517,24 @@ class PGPMIME(KryptoMIME):
             encmail.attach(submsg)
         return encmail, result
 
+def find_gnupg_key(gpg,addr,secret=False):
+    """find keyid for email 'addr' or return None.
+    If addr is a list or tuple, return a dict(addr:keyid) """
+    import email.utils
+    if type(addr) in (list,tuple):
+        result = {}
+        for a in addr: result[a] = find_gnupg_key(gpg,a,secret)
+        return result
+    addr = email.utils.parseaddr(addr)[1]
+    if not addr: return None
+    for key in gpg.list_keys(secret):
+        for uid in key['uids']:
+            if uid.find(addr)>=0:
+                return key['keyid']
+    return None
+
 class GPGMIME(PGPMIME):
-    "A PGP implementation based on GPG and the gnupg module"
+    "A PGP implementation based on GNUPG and the gnupg module"
     
     def __init__(self, gpg, default_key=None):
         "initialize with a gnupg instance and optionally a default_key (keyid,passphrase) tuple"
@@ -537,37 +559,21 @@ class GPGMIME(PGPMIME):
                 defkey,passphrase = self.default_key
             else:
                 defkey,passphrase = self.default_key,None
-            if key==True:
-                kwargs['default_key'] = defkey
-            elif key != defkey:
-                return kwargs
+            if key!=True:
+                key = self.find_key(key,secret=True)
+                if key != defkey: return kwargs
+            kwargs['default_key'] = defkey
             if not 'passphrase' in kwargs and passphrase:
                 kwargs['passphrase'] = passphrase
         return kwargs
 
     def find_key(self,addr,secret=False):
-        """find key (fingerprint) for email 'addr' or return None.
+        """find keyid for email 'addr' or return None.
         If addr is a list or tuple, return a dict(addr:keyid) """
-        import email.utils
-        if type(addr) in (list,tuple):
-            result = {}
-            for a in addr: result[a] = self.find_key(a,secret)
-            return result
-        addr = email.utils.parseaddr(addr)[1]
-        if not addr: return None
-        for key in self.gpg.list_keys(secret):
-            for uid in key['uids']:
-                if uid.find(addr)>=0:
-                    return key['keyid']
-        return None
+        return find_gnupg_key(self.gpg,addr,secret)
 
     def _sign_params(self, kwargs):
-        if 'default_key' in kwargs:
-            defkey = kwargs['default_key']
-            if type(defkey)!=bool:
-                defkey = self.find_key(defkey,secret=True)
-                if defkey: kwargs['default_key'] = defkey
-        elif self.default_key:
+        if self.default_key and not 'default_key' in kwargs:
             kwargs['default_key'] = True
         return self._set_default_key(kwargs)
 
@@ -582,37 +588,37 @@ class GPGMIME(PGPMIME):
 
     def verify_str(self, data, signature=None):
         import gnupg
-        f = gnupg._util._make_binary_stream(data, self.gpg._encoding)
         if signature:
             import os, tempfile
             tmp = tempfile.NamedTemporaryFile(prefix='gnupg',delete=False)
-            fd = tmp.file
-            fn = tmp.name
+            fd, f = tmp.file, tmp.name
+            fd.write(data)
+            fd.close()
+            tmp = tempfile.NamedTemporaryFile(prefix='gnupg',delete=False)
+            fd, fn = tmp.file, tmp.name
             fd.write(signature)
             fd.close()
             try: result = self.gpg.verify_file(f,fn)
-            finally: os.unlink(fn)
+            finally: 
+                os.unlink(f)
+                os.unlink(fn)
         else:
+            f = gnupg._util._make_binary_stream(data, self.gpg._encoding)
             result = self.gpg.verify_file(f)
-        f.close()
+            f.close()
         return result
 
     def _encrypt_params(self, recipients, kwargs):
         if 'sign' in kwargs:
             if kwargs['sign']: kwargs['default_key'] = kwargs['sign']
             del kwargs['sign']
-        if 'default_key' in kwargs:
-            defkey = kwargs['default_key']
-            if type(defkey)!=bool:
-                defkey = self.find_key(defkey,secret=True)
-                if defkey: kwargs['default_key'] = defkey
         kwargs = self._set_default_key(kwargs)
-        fingerprints = []
+        keyids = []
         for recipient in recipients:
             key = self.find_key(recipient)
-            if key: fingerprints.append(key)
-            else: fingerprints.append(recipient)
-        return fingerprints, kwargs
+            if key: keyids.append(key)
+            else: keyids.append(recipient)
+        return keyids, kwargs
 
     def encrypt_file(self, file, recipients, **kwargs):
         recipients, kwargs = self._encrypt_params(recipients, kwargs)
@@ -623,7 +629,11 @@ class GPGMIME(PGPMIME):
         return self.gpg.encrypt(data, *recipients, **kwargs)
 
     def decrypt_file(self, file, **kwargs):
-        return self.gpg.decrypt_file(file, **self._set_default_key(kwargs))
+        kwargs = self._set_default_key(kwargs)
+        if 'default_key' in kwargs: del kwargs['default_key']
+        return self.gpg.decrypt_file(file, **kwargs)
     
     def decrypt_str(self, data, **kwargs):
-        return self.gpg.decrypt(data, **self._set_default_key(kwargs))
+        kwargs = self._set_default_key(kwargs)
+        if 'default_key' in kwargs: del kwargs['default_key']
+        return self.gpg.decrypt(data, **kwargs)
